#!/bin/sh
# ============================================================================
#
# This file is part of the 'FFP-stick for ZyXEL NSA-nnn'
# http://zyxel.nas-central.org/wiki/FFP-stick
#
# This program is free software; you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# Author: Mijzelf <Mijzelf@live.com>
#
# ============================================================================
# 
# This script is supposed to be started by the firmware, due to the right
# combination of nsannn_check_file and salted_md5sum
#
# If you change anything in this file, you'll have to recreate the 
# salted_md5sum file:
# cat /etc/Zy_Private usb_key_func.sh | md5sum > salted_md5sum
# or for the NSA-3x0:
# cat /lib/libzy.so usb_key_func.sh | md5sum > salted_md5sum_3x0
#

SCRIPTVERSION="20101109"
LOGFILE="ffpboot.log"
WHILEBOOTING="while_booting.sh"
AFTERBOOTING="after_booting.sh"
TMPSTICK="/mnt/ffpstick"
TMPROOT="/mnt/ffproot"
BACKGOUND_POLLING=background_polling
REPARTITION=repartition
LOOP="0 1 2 3 4 5 6 7 8 9"

# ===================================================
# function SplitPath
# split path in first parameter in path and filename
SPLITTEDPATH="" # Output path
SPLITTEDFILE="" # Output file
# ===================================================
SplitPath()
{
    SPLITTEDPATH=""
    SPLITTEDFILE=""
    IFSBACKUP=${IFS}; 	IFS=/
    for chunk in $1
    do
        newpath=""
        if [ "${SPLITTEDPATH}" = "" ] ; then
	    newpath="${chunk}"
	else
	    newpath="${SPLITTEDPATH}/${chunk}"
	fi
	if [ "${newpath}" = "$1" ] ; then
	    SPLITTEDFILE="${chunk}"
	elif [ "/${newpath}" = "$1" ] ; then
	    SPLITTEDPATH="/${SPLITTEDPATH}"
	    SPLITTEDFILE="${chunk}"
	else
	    SPLITTEDPATH="${newpath}"
	fi
    done
    IFS=${IFSBACKUP}
    if [ "" = "${SPLITTEDPATH}" ]; then
	SPLITTEDPATH="."
    fi
}

# ===================================================
# function GetNasType
# Try to find out the NAS type, and FW
NASTYPE="" # output NAS type
NASFW="" # output FW
# Actually from the viewpoint of the FFP stick there are only 2 types:
# - NSA220 firmware 2.20 or lower, which has a different mountpoint
# - All other NASsen/firmwares
# ===================================================
GetNasType()
{
    if [ -x /zyxel/sbin/mrd_model ]; then
	model=`/zyxel/sbin/mrd_model -p`
    
	case ${model} in
	    "DC01")
		NASTYPE="NSA221"
	    ;;
	    "D401")
		NASTYPE="NSA210"
	    ;;
	    "D501")
		NASTYPE="NSA220"
	    ;;
	    "DA01")
		NASTYPE="NSA310"
	    ;;
	    "DD01")
		NASTYPE="NSA320"
	    ;;
	esac
    fi
    if [ "" = "${NASTYPE}" ]; then
	case `uname -r` in
	    "2.6.18.6"*)
		NASTYPE="NSA220"
	    ;;
	    "2.6.24.4"*)
		NASTYPE="NSA210"
	    ;;
	    *)
		NASTYPE="Unknown-${model}"
	    ;;
	esac
    fi
    if [ "NSA220" = "${NASTYPE}" ]; then
	if cat ${SPLITTEDPATH}/nsa220_check_file | grep parnerkey >/dev/null ; then
	    NASFW="230+"
	else
	    NASFW="220-"
	    return 0
	fi
    fi
    
    if [ -r /zyxel/mnt/info/fwversion ] ; then
	NASFW=` cat /zyxel/mnt/info/fwversion `
    fi
    
    if [ "" = "${NASFW}" ]; then
	NASFW="Undetermined"
    fi
}

# =========================================
# function FirstCall
# This is the original call of the script
# =========================================
FirstCall()
{
    echo ffpstick started at ` date `
    echo Try to determine NAS type...
    GetNasType
    echo "Type ${NASTYPE}, fw ${NASFW}"
    
    mountpoint="/mnt/parnerkey"
    if [ "${NASTYPE}" = "NSA220" ] && [ "${NASFW}" = "220-" ]; then
	mountpoint="/mnt"
    fi

    if [ "${mountpoint}" != "${SPLITTEDPATH}" ]; then
	echo FATAL Wrong mountpoint ${SPLITTEDPATH}, should have been ${mountpoint}
	echo Cowardly refusing to start FFP in an unknown environment
	exit 1
    fi

    if [ -e /tmp/${SPLITTEDFILE} ]; then
	echo For some reason this script is called multiple times
	exit 1
    fi

    echo Find the current usb device...
    usbdev=`cat /proc/mounts | grep ${SPLITTEDPATH} | grep -v ram | tail -n 1 | awk '{print $1}'`
    if [ "${usbdev}" = "" ]; then
	echo FATAL Usbdevice not found
	exit 1
    fi
    echo Current usb device is ${usbdev}

    # Execute an eventual '${WHILEBOOTING}' script
    if [ "${WHILEBOOTING}" != "" ] && [ -x ${SPLITTEDPATH}/${WHILEBOOTING} ]; then
	echo Executing ${WHILEBOOTING}...
	cd ${SPLITTEDPATH}
	${SPLITTEDPATH}/${WHILEBOOTING} ${usbdev} ${SPLITTEDPATH} ${NASTYPE} "${NASFW}"
    fi
    
    COMMAND=${BACKGOUND_POLLING}

    if [ -x ${SPLITTEDPATH}/${AFTERBOOTING} ]; then
	echo Check for filesystem...
	filesystem=`cat /proc/mounts | grep ${SPLITTEDPATH} | grep -v ram | tail -n 1 | awk '{print $3}'`
	echo Filesystem is ${filesystem}
	if [ "${filesystem}" = "vfat" ] ; then
	    maindev=` echo ${usbdev} | sed 's/[0-9]*//g' `
	    if [ "${maindev}" = "${usbdev}" ]; then
		echo ${usbdev} is not partitioned. Will have to partition it
		COMMAND=${REPARTITION}
	    else
		echo Check ${maindev} for number of partitions...
	        partition_num=`fdisk -l ${maindev} | grep "^"${maindev} | wc -l `
		echo ${maindev} has ${partition_num} partitions
		if [ ${partition_num} -eq 1 ] ; then
		    echo Will have to repartition
		    COMMAND=${REPARTITION}
		    usbdev=${maindev}
		fi
	    fi
	fi
    else
	echo No ${AFTERBOOTING} in the root of the ffpstick found
	echo This is not the first run, or FFP is supposed to run from a different place
    fi

    if [ "${COMMAND}" = "${REPARTITION}" ] ; then
	echo Copy files to ramdisk...
	filelist="nsa210_check_file md5sum \
		nsa220_check_file \
		NSA221_check_file \
		nsa310_check_file_C0 salted_md5sum_3x0 \
		salted_md5sum"
		
	copied=""
	notcopied=""
	for savedir in ${TMPSTICK} ${TMPROOT}
	do
	    if ! [ -d ${savedir} ]; then
		mkdir ${savedir}
	    fi
	    for file in ${filelist} 
	    do
		if [ -r ${SPLITTEDPATH}/${file} ] ; then
		    if cp ${SPLITTEDPATH}/${file} ${savedir}/${file}; then
			copied="${copied} ${file}"
		    else
			rm ${savedir}/${file}
			notcopied="${notcopied} ${file}"
		    fi
		fi
	    done
	    filelist="${AFTERBOOTING} rootfs.tgz fun_plug.tgz"
	done
	echo "Copied ${copied}"
	if [ "" != "${notcopied}" ]; then
	    echo "*Not* copied ${notcopied}"
	fi
    fi

    echo Copy myself to /tmp...
    cp -pf ${SPLITTEDPATH}/${SPLITTEDFILE} /tmp/${SPLITTEDFILE}

    if ! [ -x /tmp/${SPLITTEDFILE} ]; then
	echo FATAL Failed to copy!
	exit 1
    fi

    COMMAND="/tmp/${SPLITTEDFILE} ${COMMAND} ${usbdev} ${NASTYPE} \"${NASFW}\""
    echo And execute...
    echo "${COMMAND} &"

    if [ "" != "${LOGFILE}" ]; then
	cp -a ${SPLITTEDPATH}/${LOGFILE} /tmp/${LOGFILE}
	# Close logfile, maybe it will disturb the unmounting hereafter
        exec >/dev/null 2>&1
        sync
    fi
    cd /tmp
    
    ${COMMAND} &
    exit 1
}

# ==========================================================================
# function ReopenLogfile
# Arguments: device mountpoint copy/movecommand
# Mounts device on mountpoint, moves/copies current logfile to mountpoint
# and redirects output
# ==========================================================================
ReopenLogfile()
{
    if [ "" = "${LOGFILE}" ] ; then
	return 0
    fi

    if ! [ -d $2 ]; then
        mkdir $2
    fi
    if ! mount $1 $2 ; then
	return 1
    fi
    $3 /tmp/${LOGFILE} $2/${LOGFILE}
    exec >>$2/${LOGFILE} 2>&1
}

# ================================================
# function ProbeMountPoints
# Check all mountpoints for the existence of
# - usb_key_func.sh (this file) being the ffpstick
# - ffproot/after_booting.sh being the ffproot
# Outputs:
MP_FFPSTICK=""
MP_FFPROOT=""
#=================================================
ProbeMountPoints()
{
    echo Probe all mount points $1 seconds...

    mountpoints=`cat /proc/mounts | awk '{print $2}'`

    for mp in ${mountpoints}
    do
	# count number of / in string. We're only interested with 2 or more
	slashes=` echo ${mp} | tr -cd / `
	if [ ${#slashes} -gt 1 ]; then
	    echo probe ${mp}...
	    if [ "" = "${MP_FFPROOT}" ] && [ -x ${mp}/ffproot/${AFTERBOOTING} ] ; then
		MP_FFPROOT=${mp}/ffproot
		echo found ffproot on ${MP_FFPROOT}
	    fi

	    if [ "" = "${MP_FFPSTICK}" ] ; then #&& [ -x ${mp}/${SPLITTEDFILE} ] ; then
		if cat /proc/mounts | grep ${mp} | grep $2 >/dev/null ; then
		    MP_FFPSTICK=${mp}
		    if [ "" != "${LOGFILE}" ] ; then
			echo "found ffpstick on ${MP_FFPSTICK}, moving logfile"
			mv ${SPLITTEDPATH}/${LOGFILE} ${MP_FFPSTICK}/${LOGFILE}
			exec >>${MP_FFPSTICK}/${LOGFILE} 2>&1
		    fi
		fi
	    fi
	fi
    done

    if [ "" = "${MP_FFPROOT}" ]; then
        echo No /anymountpoint/ffproot/${AFTERBOOTING} found
        return 1
    fi
    return 0
} 

# ===========================================
# function BackgroundPolling
# function waits until NAS is booted, to start ffp
# ===========================================
BackgroundPolling()
{
    echo Wait for the stick to be mounted again by the firmware...
    sleep 30

    if [ -r ${SPLITTEDPATH}/${SPLITTEDFILE} ]; then
	rm ${SPLITTEDPATH}/${SPLITTEDFILE}
    fi
    
    ffpstickfound=0
    for outer in ${LOOP}
    do
	for inner in ${LOOP}
        do
    	    ProbeMountPoints ${outer}${inner}0 $1
    	    if [ "" != "${MP_FFPROOT}" ] ; then
	    	cd ${MP_FFPROOT}
		exec ${MP_FFPROOT}/${AFTERBOOTING} ${MP_FFPROOT} ${NASTYPE} "${NASFW}"
	        exit 0
    	    fi
    	    if [ "" != "${MP_FFPSTICK}" ] ; then 
    		if [ ${ffpstickfound} -eq 3 ] ; then
    		    # Have a look if we can mount the 2nd partition ourselfs
    		    maindev=` echo $1 | sed 's/[0-9]*//g' `
		    if [ "${maindev}" != "$1" ] ; then
        		echo "30 seconds should be enough for anybody ;)"
			echo Will try to mount ffproot myself
			mkdir ${TMPROOT}
			mount ${maindev}2 ${TMPROOT}
		    fi
    		fi
    		ffpstickfound=`expr ${ffpstickfound} + 1`
    	    fi
	    sleep 10
        done
    done

    # Have a look if we can mount the 2nd partition ourselfs
    maindev=` echo $1 | sed 's/[0-9]*//g' `
    if [ "${maindev}" != "$1" ] ; then
	ReopenLogile $1 ${TMPSTICK} mv
	echo Final attempt to mount the stick myself
	mkdir ${TMPROOT}
	if mount ${maindev}2 ${TMPROOT} ; then
	    MP_FFPROOT=${TMPROOT}/ffproot
	    if [ -x ${MP_FFPROOT}/${AFTERBOOTING} ] ; then
	    	cd ${MP_FFPROOT}
		exec ${MP_FFPROOT}/${AFTERBOOTING} ${MP_FFPROOT} ${NASTYPE} "${NASFW}"
	        exit 0
	    fi

	    echo FATAL No ${MP_FFPROOT}/${AFTERBOOTING} found 
	    exit 0
	fi
	
	echo FATAL Mounting failed
	exit 0
    fi

    echo "FATAL Timeout & Giving up"
    ReopenLogfile $1 ${TMPSTICK} mv
    exit 0
}

# =======================================
# function WaitForUnmount
# is called before attempting to repartition
# It returns when the stick is umounted,
# or on a timeout (100 sec)
# =======================================
WaitForUnmount()
{
    for outer in ${LOOP}
    do
	for inner in ${LOOP}
        do
	    if ! cat /proc/mounts | grep ${device} >/dev/null ; then
		return 0
	    fi
	    echo ${outer}${inner} seconds...
	    sleep 1
        done
    done
    return 1
}

# ==============================================
# function ForceUmount
# is called when a timeout occurs in waiting for umounting
# of the ffpstick
# ==============================================
ForceUmount()
{
    # Try to umount intermittently on devicename and mountpoint
    for outer in ${LOOP}
	do
	for inner in d m
	do
	    if [ "${inner}" = "d" ] ; then
		echo umount $1
		umount $1
	    else
	        mountpoints=`cat /proc/mounts | grep $1 | grep -v ram | awk '{print $2}'`
		for m in ${mountpoints}
		do
		    echo umount ${m}
		    umount ${m}
		done
	    fi
	    
	    if ! cat /proc/mounts | grep $1 >/dev/null ; then
		# Yeah, unmounted!
		return 0
	    fi
	done
    done
    return 1
}

# ================================================
# function FatalErrorInRepartitioning
# Is called when repartitioning failed.
# It will try to put the logfile back on the stick
# and collect more info
# ================================================
FatalErrorInRepartitioning()
{
    device=$1
    echo FATAL $2
    if ! ReopenLogfile ${device}1 ${TMPSTICK} mv ; then
    	ReopenLogfile ${device} ${TMPSTICK} mv
    fi
    for dmp in mounts filesystems partitions
    do
	echo =================
	echo cat /proc/${dmp}
	cat /proc/${dmp}
    done
    echo ================
    echo dmesg
    dmesg
    # umount stick, to make it possible to automount it
    exec >/dev/null 2>&1
    umount ${TMPSTICK}
    rmdir ${TMPSTICK}
    exit 0
}

Repartition()
{
    device=$1
    echo Repartition ${device}
    echo Wait until stick is no longer mounted...
    
    # Clear dmesg, just in case we get a fatal error. 
    dmesg -c >/dev/null 2>&1
    
    WaitForUnmount $1
    if [ $? -eq 1 ] ; then
	echo Tired of waiting, will try to umount myself...
	ForceUmount ${device}
	if [ $? -eq 1 ] ; then
	    FatalErrorInRepartitioning ${device} "timeout in waiting & couldn't force umount"
	fi
    fi

    echo Unmounted at ` date `. Repartitioning...
    echo -e "o\nn\np\n1\n\n+16M\nn\np\n2\n\n\nt\n1\n4\nw\n" | fdisk ${device}
    if [ "0" != "$?" ]; then
        FatalErrorInRepartitioning ${device} "Partitioning failed"
    fi
    echo Done at ` date `. Creating and mounting filesystems...
    if ! mkdosfs ${device}1 ; then
	FatalErrorInRepartitioning ${device} "Couldn\'t create vfat fs on ${device}1"
    fi
    sync
    mkdir /tmp/mp
    mount ${device}1 /tmp/mp
    if [ "0" != "$?" ]; then
	FatalErrorInRepartitioning ${device} "Mount ${device}1 on /tmp/mp failed"
    fi
    mv ${TMPSTICK}/* /tmp/mp/
    cp ${SPLITTEDPATH}/${SPLITTEDFILE} /tmp/mp/
    umount /tmp/mp
    rmdir ${TMPSTICK}

    if ! mke2fs -j ${device}2 ; then
	FatalErrorInRepartitioning ${device} "Couldn\'t create ext3 fs on ${device}2"
    fi
    sync
    mount ${device}2 /tmp/mp
    if [ "0" != "$?" ]; then
	FatalErrorInRepartitioning ${device} "Mount ${device}2 on /tmp/mp failed"
    fi

    mkdir /tmp/mp/ffproot
    mv ${TMPROOT}/* /tmp/mp/ffproot/
    umount /tmp/mp
    rmdir ${TMPROOT}
    rmdir /tmp/mp

    echo Done at ` date `.

    echo Call BackgroundPolling...
    ReopenLogfile ${device}1 ${TMPSTICK} cp
    exec >>${SPLITTEDPATH}/${LOGFILE} 2>&1
    umount ${TMPSTICK}
    rmdir ${TMPSTICK}

    BackgroundPolling ${device}1
}

# Find filename and path
SplitPath $0

if [ "" != "${LOGFILE}" ]; then
    logfile=${SPLITTEDPATH}/${LOGFILE}
    if [ $# -eq 0 ]; then
	templog=${SPLITTEDPATH}/templog.log
	exec >>${templog} 2>&1
	# Try to save older bootlogs
	echo Some logfile administration...
	oldlog=${logfile}.9.gz
	if [ -r ${oldlog} ] ; then
	    echo rm ${oldlog}
	    rm ${oldlog}
	fi
	for nr in 8 7 6 5 4 3 2 1
	do
	    oldlog=${logfile}.${nr}.gz
	    if [ -r ${oldlog} ] ; then
		newnr=`expr $nr + 1`
		newlog=${logfile}.${newnr}.gz
		echo mv ${oldlog} ${newlog}
		mv ${oldlog} ${newlog}
	    fi
	done
	if [ -r ${logfile} ] ; then
	    echo "cat ${logfile} | gzip -c >${oldlog}"
	    cat ${logfile} | gzip -c >${oldlog}
	    rm ${logfile}
	fi
	exec >/dev/null 2>&1
	mv ${templog} ${logfile}
    fi
    exec >>${logfile} 2>&1
fi

echo Script ${SPLITTEDFILE} version ${SCRIPTVERSION} running from ${SPLITTEDPATH}

if [ $# -gt 1 ]; then
    NASTYPE=$3
    NASFW=$4

    if [ "$1" = "${REPARTITION}" ]; then
	Repartition $2
	exit 0
    fi
    if [ "$1" = "${BACKGOUND_POLLING}" ]; then
	BackgroundPolling $2 ""
	exit 0
    fi
    if [ "$1" = "NSA221Finish" ]; then
	umount /mnt/parnerkey
	rmdir /mnt/parnerkey
	if [ -x /tmp/NSA221Finish.sh ]; then
	    exec /tmp/NSA221Finish.sh
	fi
	exit 0
    fi
    echo "Meuh?"
    exit 0
fi

FirstCall
exit 1
